# coding: utf-8

import json
import six
from copy import deepcopy

from django import forms
from django.conf import settings
from django.db import transaction

from yaat.models import Column


class HeadersField(forms.Field):
    def to_python(self, value):
        if isinstance(value, six.string_types):
            try:
                return json.loads(value)
            except ValueError:
                raise forms.ValidationError('Enter valid JSON')
        return value


class YaatValidatorForm(forms.Form):
    limit = forms.ChoiceField()
    offset = forms.IntegerField(min_value=0, initial=1)
    headers = HeadersField(required=False)

    def __init__(self, *args, request, columns, resource, **kwargs):
        self.resource = resource
        self.request = request
        self.user = getattr(self.request, getattr(settings, 'YAAT_REQUEST_ATTR', 'user'))
        self.columns = columns
        self.stateful_init = self.resource._meta.stateful_init
        self._column_fields = {column.key: i for i, column in enumerate(self.columns)}

        super().__init__(*args, **kwargs)

        if self.resource._meta.limit is None:
            self.fields['limit'] = forms.IntegerField(min_value=0, initial=10)
        else:
            self.fields['limit'].initial = self.resource._meta.limit
            self.fields['limit'].choices = [(_, _) for _ in self.resource._meta.limit_choices]

    def _get_column(self, name):
        retval = deepcopy(self.columns[self._column_fields[name]])
        retval.pk = None
        return retval

    def full_clean(self):
        super().full_clean()
        is_init = False

        if 'limit' in self._errors.keys():
            self.cleaned_data['limit'] = self.fields['limit'].initial
            del self._errors['limit']

        if 'offset' in self._errors.keys():
            self.cleaned_data['offset'] = self.fields['offset'].initial
            del self._errors['offset']

        if 'headers' in self._errors.keys():
            is_init = True
            self.cleaned_data['headers'] = self.columns
            del self._errors['headers']

        if self.stateful_init:
            session_key = self.session_key(self.resource)
            state = {
                'limit': int(self.cleaned_data['limit']),
                'offset': self.cleaned_data['offset']
            }

            if is_init:
                state = self.request.session.get(session_key, state)
                self.cleaned_data['limit'] = state['limit']
                self.cleaned_data['offset'] = state['offset']
            self.request.session[session_key] = state

    @classmethod
    def limit_dict(cls, request, resource):
        if resource._meta.stateful_init:
            state = request.session.get(cls.session_key(resource), None)
            if state:
                limit = state['limit']
            else:
                limit = resource._meta.limit
        else:
            limit = resource._meta.limit

        return {
            'limit': limit,
            'options': resource._meta.limit_choices
        }

    def invalidate_state(self):
        self.request.session.pop(self.session_key(self.resource), None)

    def reset_offset(self):
        self.cleaned_data['offset'] = self.fields['offset'].initial
        if self.stateful_init:
            session_key = self.session_key(self.resource)
            state = self.request.session.get(session_key, None)
            if state:
                state['offset'] = self.cleaned_data['offset']
                self.request.session[session_key] = state

    @staticmethod
    def session_key(resource):
        return 'yaat_init_state_' + resource._meta.resource_name

    def clean_headers(self):
        posted = self.cleaned_data['headers']

        headers = []

        try:
            for head in posted:
                try:
                    col = self._get_column(head['key'])
                    headers.append(col)

                    if col.ordering != Column.ORDER_DISALLOWED:
                        col.ordering = head['order']

                    if col.is_shown != Column.HIDE_DISALLOWED:
                        col.is_shown = not head['hidden']
                except KeyError as k:
                    raise forms.ValidationError('Missing property {0!r}'.format(k))
        except TypeError:
            raise forms.ValidationError('Missing headers')

        return headers

    def save(self):
        if self.columns == self.cleaned_data['headers']:
            return

        for i in range(0, len(self.cleaned_data['headers'])):
            self.cleaned_data['headers'][i].order = i + 1
            self.cleaned_data['headers'][i].user = self.user

        with transaction.atomic():
            col = self.cleaned_data['headers'][0]
            Column.objects.filter(user_id=col.user_id, resource=col.resource).delete()
            Column.objects.bulk_create(self.cleaned_data['headers'])
            Column.cached.bulk_create(self.cleaned_data['headers'])



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        ('yaat', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='column',
            name='user',
            field=models.ForeignKey(to=getattr(settings, 'YAAT_FOREIGN_KEY', settings.AUTH_USER_MODEL), verbose_name='User', related_name='columns', on_delete=models.CASCADE),
        ),
    ]



# coding: utf-8

from restify.resource.model import ModelResourceOptions
from restify.resource.base import ResourceMeta
from restify.serializers import DjangoSerializer


class YaatModelResourceOptions(ModelResourceOptions):
    serializer = DjangoSerializer
    columns = ()
    stateful = False
    stateful_init = False
    limit = None
    limit_choices = [limit]


class YaatModelResourceMeta(ResourceMeta):
    options_class = YaatModelResourceOptions



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Column',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('order', models.PositiveIntegerField(editable=False, db_index=True)),
                ('resource', models.CharField(max_length=64, verbose_name='Resource name')),
                ('key', models.CharField(max_length=64, verbose_name='Column key')),
                ('is_shown', models.NullBooleanField(verbose_name='Show field', default=True)),
                ('ordering', models.PositiveSmallIntegerField(choices=[(None, 'Ordering disallowed'), (0, 'Unordered'), (1, 'Ascending'), (2, 'Descending')], null=True, verbose_name='Field order', default=0)),
                ('is_virtual', models.BooleanField(default=True)),
                ('user', models.ForeignKey(to=settings.AUTH_USER_MODEL, verbose_name='User', related_name='column_users', on_delete=models.CASCADE)),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='column',
            unique_together=set([('resource', 'user', 'key')]),
        ),
    ]



# coding: utf-8

from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.core.cache import cache
from django.core.paginator import Paginator, EmptyPage
from django.apps import apps
from restify.http import status
from restify.http.response import ApiResponse
from restify.resource import Resource
from restify.resource.model import ModelResourceMixin

from .forms import YaatValidatorForm
from .meta import YaatModelResourceMeta
from .models import Column


class YaatModelResource(Resource, ModelResourceMixin, metaclass=YaatModelResourceMeta):
    VALIDATOR_FORM = YaatValidatorForm

    @classmethod
    def invalidate_column_cache(cls, user):
        user_class = apps.get_model(getattr(settings, 'YAAT_FOREIGN_KEY', settings.AUTH_USER_MODEL))
        if not isinstance(user, user_class):
            raise TypeError('expected {0!r} received {1!r} instead'.format(user_class, user.__class__))

        key = Column.cached.KEY_TEMPLATE % (user.pk, cls._meta.resource_name)
        cache.delete(key)

    def get_columns(self):
        columns = []

        for c in self._meta.columns:
            if isinstance(c, str):
                field = self._meta.model._meta.get_field(c)
                column = Column(key=field.name,
                                value=field.verbose_name,
                                resource=self._meta.resource_name,
                                is_virtual=False)
                columns.append(column)
            elif isinstance(c, Column):
                c.resource = self._meta.resource_name
                columns.append(c)
            else:
                raise TypeError('Column item is expected to be string or Column')

        return columns

    def get_stateful_columns(self, columns):
        retval = []

        mapper = {columns[i].key: i for i in range(0, len(columns))}

        if self._meta.stateful and not isinstance(self.request.user, AnonymousUser):
            column_user = getattr(self.request, getattr(settings, 'YAAT_REQUEST_ATTR', 'user'))
            retval = Column.cached.filter(resource=self._meta.resource_name, user=column_user)

        for col in retval:
            col.value = columns[mapper[col.key]].value

        return retval

    def get_queryset_order_keys(self, columns):
        keys = []
        for column in columns:
            ordering = column.get_ordering()
            if ordering:
                keys.append(ordering)
        return keys

    def get_queryset(self, columns):
        order_keys = self.get_queryset_order_keys(columns)
        return super().get_queryset().order_by(*order_keys).all()

    def get_page(self, qs, limit, page_number=1):
        paginator = Paginator(qs, limit)
        return paginator.page(page_number)

    def row_hook(self, row, obj):
        return row

    def get_rows(self, page, cols, **kwargs):
        rows = []
        for obj in page:
            cells = []
            for col in cols:
                if col.is_shown != False:
                    value = getattr(obj, col.key)
                    if hasattr(value, '__call__'):
                        value = value(**kwargs)
                    cells.append(value)
            rows.append(self.row_hook({'id': obj.pk, 'values': cells}, obj))
        return rows

    def common(self, request, *args, **kwargs):
        available_columns = self.get_columns()
        stateful_columns = self.get_stateful_columns(available_columns)

        self.validator_form = self.VALIDATOR_FORM(
            request.POST,
            request=request,
            columns=stateful_columns or available_columns,
            resource=self)

    def _compose(self, form):
        queryset = self.get_queryset(form.cleaned_data['headers'])
        page = self.get_page(queryset, form.cleaned_data['limit'], form.cleaned_data['offset'])
        rows = self.get_rows(page, form.cleaned_data['headers'])
        return page, rows

    def post(self, request, *args, **kwargs):
        form = self.validator_form
        if form.is_valid():
            try:
                page, rows = self._compose(form)
            except EmptyPage:
                form.reset_offset()
                page, rows = self._compose(form)

            if self._meta.stateful and not isinstance(request.user, AnonymousUser):
                form.save()

            return ApiResponse({'columns': form.cleaned_data['headers'],
                                'rows': rows,
                                'pages': page})
        else:
            return ApiResponse(form, status_code=status.HTTP_400_BAD_REQUEST)



# coding: utf-8

from setuptools import setup
from yaat import __version__

vcs_dependencies = [
    # order_with_respect_to can be a tuple in this fork
    'git+https://github.com/yuvadm/django-ordered-model@b875b891b838b26547e1b767e30badc8f4a0331e#egg=django-ordered-model'
]

setup(
    name='django-yaat',
    version=__version__,
    packages=['yaat'],
    install_requires=['Django>=2.0', 'django-restify-framework>=1.0.0', 'django-ordered-model>=1.4.3'],
    include_package_data=True,
    dependency_links=vcs_dependencies,
    description='The django-restify-framework resource of the yaat AngularJs module',
    author='Slapec',
    url='https://github.com/slapec/django-yaat',
    license='MIT',
    zip_safe=False,
)



# coding: utf-8

from django.conf import settings
from django.core.cache import cache
from django.db import models
from django.utils.translation import ugettext_lazy as _
from ordered_model.models import OrderedModel


class CachedManager(models.Manager):
    KEY_TEMPLATE = '%d_%s_columns'

    def filter(self, resource, user):
        key = self.KEY_TEMPLATE % (user.pk, resource)

        value = cache.get(key)

        if value is None:
            value = list(Column.objects.filter(resource=resource, user=user))
            cache.set(key, value)

        return value

    def bulk_create(self, columns):
        key = self.KEY_TEMPLATE % (columns[0].user_id, columns[0].resource)
        cache.set(key, columns)


class Column(OrderedModel):
    ORDER_DISALLOWED = None
    UNORDERED = 0
    ASC = 1
    DESC = 2

    HIDE_DISALLOWED = None

    ORDER_CHOICES = (
        (ORDER_DISALLOWED, _('Ordering disallowed')),
        (UNORDERED, _('Unordered')),
        (ASC, _('Ascending')),
        (DESC, _('Descending'))
    )

    user = models.ForeignKey(
        getattr(settings, 'YAAT_FOREIGN_KEY', settings.AUTH_USER_MODEL),
        verbose_name=_('User'),
        related_name='columns',
        on_delete=models.CASCADE
    )

    resource = models.CharField(max_length=64, verbose_name=_('Resource name'))
    key = models.CharField(max_length=64, verbose_name=_('Column key'))
    is_shown = models.NullBooleanField(default=True, verbose_name=_('Show field'))
    ordering = models.PositiveSmallIntegerField(choices=ORDER_CHOICES, default=UNORDERED, null=True,
                                                verbose_name=_('Field order'))
    is_virtual = models.BooleanField(default=True)

    order_with_respect_to = ('resource', 'user')

    objects = models.Manager()
    cached = CachedManager()

    class Meta:
        unique_together = ('resource', 'user', 'key')

    def __init__(self, *args, **kwargs):
        self.value = kwargs.pop('value', None)
        super().__init__(*args, **kwargs)

        if self.is_virtual:
            self.ordering = self.ORDER_DISALLOWED

    def __eq__(self, other):
        left = [self.user_id, self.resource, self.key, self.order, self.is_shown, self.ordering]
        right = [other.user_id, other.resource, other.key, other.order, other.is_shown, other.ordering]

        return left == right

    def get_ordering(self):
        if self.is_shown is False:
            return None

        if self.ordering == self.ASC:
            return self.key
        elif self.ordering == self.DESC:
            return '-' + self.key

    def flatten(self):
        if self.value is None:
            raise ValueError('Cannot flatten because self.value is None')

        data = {
            'key': self.key,
            'value': str(self.value)
        }

        if self.ordering != self.ORDER_DISALLOWED:
            data['order'] = self.ordering

        if self.is_shown is not None:
            data['hidden'] = not self.is_shown

        return data
